@@ -1,32 +1,36 @@
-from requests.packages.urllib3.exceptions import InsecureRequestWarning
+
+from foundry.xtract_method import *
+import time
+import h5py
+import glob
+import json
+import requests
+import mdf_toolbox
+from json2table import convert
+import pandas as pd
+from datetime import date
+from typing import Any
+import multiprocessing
+from mdf_connect_client import MDFConnectClient
+from mdf_forge import Forge
+from dlhub_sdk.utils.schemas import validate_against_dlhub_schema
+from dlhub_sdk.models.servables.keras import KerasModel
+from dlhub_sdk.models.servables.sklearn import ScikitLearnModel
+from dlhub_sdk import DLHubClient
+from collections import namedtuple
+from joblib import Parallel, delayed
+from pydantic import AnyUrl, ValidationError
 from foundry.models import (
     FoundryMetadata,
     FoundryConfig,
     FoundrySpecificationDataset,
     FoundrySpecification,
 )
-from pydantic import AnyUrl, ValidationError
-from joblib import Parallel, delayed
-from collections import namedtuple
-from dlhub_sdk import DLHubClient
-# TODO: do imports nicer
-from dlhub_sdk.models.servables.sklearn import ScikitLearnModel
-from dlhub_sdk.models.servables.keras import KerasModel
-from dlhub_sdk.utils.schemas import validate_against_dlhub_schema
-from mdf_forge import Forge
-from mdf_connect_client import MDFConnectClient
-import multiprocessing
-from typing import Any
-from datetime import date
-import multiprocessing
-import pandas as pd
-import mdf_toolbox
-import requests
-import json
-import glob
-import h5py
-import time
+from requests.packages.urllib3.exceptions import InsecureRequestWarning
+import logging
 import os
+import shutil
+logging.disable(logging.INFO)
 
 
 class Foundry(FoundryMetadata):
@@ -45,28 +49,32 @@ class Foundry(FoundryMetadata):
     xtract_tokens: Any
 
     def __init__(
-        self, no_browser=False, no_local_server=False, search_index="mdf-test", **data
+        self, no_browser=False, no_local_server=False, index="mdf-test", authorizers=None, **data
     ):
         super().__init__(**data)
-        auths = mdf_toolbox.login(
-            services=[
-                "data_mdf",
-                "mdf_connect",
-                "search",
-                "petrel",
-                "transfer",
-                "dlhub",
-                "openid",
-                "https://auth.globus.org/scopes/facd7ccc-c5f4-42aa-916b-a0e270e2c2a9/all",
-            ],
-            app_name="Foundry",
-            make_clients=True,
-            no_browser=no_browser,
-            no_local_server=no_local_server,
-        )
+
+        if authorizers:
+            auths = authorizers
+        else:
+            auths = mdf_toolbox.login(
+                services=[
+                    "data_mdf",
+                    "mdf_connect",
+                    "search",
+                    "petrel",
+                    "transfer",
+                    "dlhub",
+                    "openid",
+                    "https://auth.globus.org/scopes/facd7ccc-c5f4-42aa-916b-a0e270e2c2a9/all",
+                ],
+                app_name="Foundry",
+                make_clients=True,
+                no_browser=no_browser,
+                no_local_server=no_local_server,
+            )
 
         self.forge_client = Forge(
-            index=search_index,
+            index=index,
             services=None,
             search_client=auths["search"],
             transfer_client=auths["transfer"],
@@ -74,9 +82,14 @@ class Foundry(FoundryMetadata):
             petrel_authorizer=auths["petrel"],
         )
 
+        if index == "mdf":
+            test = False
+        else:
+            test = True
         # TODO: when release-ready, remove test=True
+
         self.connect_client = MDFConnectClient(
-            authorizer=auths["mdf_connect"], test=True
+            authorizer=auths["mdf_connect"], test=test
         )
 
         self.dlhub_client = DLHubClient(
@@ -85,9 +98,12 @@ class Foundry(FoundryMetadata):
             fx_authorizer=auths[
                 "https://auth.globus.org/scopes/facd7ccc-c5f4-42aa-916b-a0e270e2c2a9/all"
             ],
+            openid_authorizer=auths['openid'],
             force_login=False,
         )
 
+
+
         self.xtract_tokens = {
             "auth_token": auths["petrel"].access_token,
             "transfer_token": auths["transfer"].authorizer.access_token,
@@ -96,12 +112,15 @@ class Foundry(FoundryMetadata):
             ].access_token,
         }
 
-    def load(self, name, version="1.1", provider="MDF", download=True, globus=True, verbose=False, **kwargs,):
+    def load(self, name, download=True, globus=True, verbose=False, metadata=None, authorizers=None, **kwargs):
         """Load the metadata for a Foundry dataset into the client
         Args:
             name (str): Name of the foundry dataset
             download (bool): If True, download the data associated with the package (default is True)
-    
+            globus (bool): If True, download using Globus, otherwise https
+            verbose (bool): If True print additional debug information
+            metadata (dict): **For debug purposes.** A search result analog to prepopulate metadata.
+
         Keyword Args:
             interval (int): How often to poll Globus to check if transfers are complete
 
@@ -109,42 +128,44 @@ class Foundry(FoundryMetadata):
         -------
             self
         """
-        res = []
-        # MDF specific logic
-        # Handle DOI inputs
-        if name.startswith("10.") and provider == "MDF":
-            print("Loading by DOI")
-            res = (
-                self.forge_client.match_dois(name)
-                .match_resource_types("dataset")
-                .match_field("mdf.organizations", "foundry")
-                .search()
-            )
 
-        # Handle MDF source_ids
-        else:
-            print("Loading by source_id")
-            res = (
-                self.forge_client.match_field("mdf.organizations", "foundry")
-                .match_resource_types("dataset")
-                .match_field("mdf.source_id", name)
-                .search()
-            )
-
-        if not res:
-            return self
-        else:
-            # res is list of all matches in Forge
-            # TODO: handle if there are multiple matches
-            res = res[0]
+        # handle empty dataset name (was returning all the datasets)
+        if not name:
+            raise ValueError("load: No dataset name is given")
             
-            # map result to FoundryMetadata Pydantic structure (see models.py)
-            res["dataset"] = res["projects"]["foundry"]
-            res["dataset"]["type"] = res["dataset"]["package_type"]
-            del res["projects"]["foundry"]
+        if metadata:
+            res = metadata
 
-            # TODO: reassign values to self in a safer way
-            self = Foundry(**res)
+        if metadata:
+            res = metadata
+
+        # MDF specific logic
+        if is_doi(name) and not metadata:
+            res = self.forge_client.match_resource_types("dataset")
+            res = res.match_dois(name).search()
+
+        else:
+            res = self.forge_client.match_field(
+                "mdf.organizations", self.config.organization
+            ).match_resource_types("dataset")
+            res = res.match_field("mdf.source_id", name).search()
+
+        # unpack res, handle if empty
+        try:
+            # if search returns multiple results, this automatically uses first result
+            res = res[0]
+        except IndexError as e:
+            raise Exception("load: No metadata found for given dataset") from e
+
+        try:
+            res["dataset"] = res["projects"][self.config.metadata_key]
+        except KeyError as e:
+            raise Exception("load: not able to index with metadata key {}".format(
+                self.config.metadata_key)) from e
+
+        del res["projects"][self.config.metadata_key]
+
+        self = Foundry(**res, authorizers=authorizers)
 
         if download is True:  # Add check for package existence
             self.download(
@@ -161,7 +182,8 @@ class Foundry(FoundryMetadata):
             (pandas.DataFrame): DataFrame with summary list of Foundry data packages including name, title, and publication year
         """
         res = (
-            self.forge_client.match_field("mdf.organizations", "foundry")
+            self.forge_client.match_field(
+                "mdf.organizations", self.config.organization)
             .match_resource_types("dataset")
             .search()
         )
@@ -213,7 +235,7 @@ class Foundry(FoundryMetadata):
 
         for package in packages:
             self = self.load(package)
-            X, y = self.load_data()
+            X, y = self.load_data()  # TODO: update how this is unpacked, out of date
             X["source"] = package
             y["source"] = package
             X_frames.append(X)
@@ -242,57 +264,51 @@ class Foundry(FoundryMetadata):
         """Load in the data associated with the prescribed dataset
 
         Tabular Data Type: Data are arranged in a standard data frame
-        stored in self.dataframe_file. The contents are read, and 
+        stored in self.dataframe_file. The contents are read, and
 
         File Data Type: <<Add desc>>
 
         For more complicated data structures, users should
-        subclass Foundry and override the load_data function 
+        subclass Foundry and override the load_data function
 
         Args:
            inputs (list): List of strings for input columns
-           outputs (list): List of strings for output columns
+           targets (list): List of strings for output columns
 
         Returns
-        -------
+        -------s
              (tuple): Tuple of X, y values
         """
+        data = {}
 
-        if source_id:
-            path = os.path.join(self.config.local_cache_dir, source_id)
-            print("Here")
+        # Handle splits if they exist. Return as a labeled dictionary of tuples
+        try:
+            if self.dataset.splits:
+                for split in self.dataset.splits:
+                    data[split.label] = self._load_data(file=split.path,
+                                                        source_id=source_id, globus=globus)
+                return data
+            else:
+                return {"data": self._load_data(source_id=source_id, globus=globus)}
+        except Exception as e:
+            raise Exception(
+                "Metadata not loaded into Foundry object, make sure to call load()") from e
+
+    def _repr_html_(self) -> str:
+        if not self.dc:
+            buf = str(self)
         else:
-            path = os.path.join(self.config.local_cache_dir, self.mdf["source_id"])
-        # Handle Foundry-defined types.
-        if self.dataset.type.value == "tabular":
-            # If the file is not local, fetch the contents with Globus
-            # Check if the contents are local
-            # TODO: Add hashes and versioning to metadata and checking to the file
-            try:
-                self.dataset.dataframe = pd.read_json(
-                    os.path.join(path, self.config.dataframe_file)
-                )
-            except:
-                # Try to read individual lines instead
-                self.dataset.dataframe = pd.read_json(
-                    os.path.join(path, self.config.dataframe_file), lines=True
-                )
+            title = self.dc['titles'][0]['title']
+            authors = [creator['creatorName']
+                       for creator in self.dc['creators']]
+            authors = '; '.join(authors)
 
-            return (
-                self.dataset.dataframe[self.dataset.inputs],
-                self.dataset.dataframe[self.dataset.outputs],
-            )
-        elif self.dataset.type.value == "hdf5":
-            f = h5py.File(os.path.join(path, self.config.data_file), "r")
-            inputs = [f[i[0:]] for i in self.dataset.inputs]
-            outputs = [f[i[0:]] for i in self.dataset.outputs]
-            return (inputs, outputs)
-        else:
-            raise NotImplementedError
+            buf = f'<h2>{title}</h2>{authors}'
 
-    def describe(self):
-        print("DC:{}".format(self.dc))
-        print("Dataset:{}".format(self.dataset.json(exclude={"dataframe"})))
+            buf = f'{buf}<h3>Dataset</h3>{convert(json.loads(self.dataset.json(exclude={"dataframe"})))}'
+        # buf = f'{buf}<h3>MDF</h3>{convert(self.mdf)}'
+        # buf = f'{buf}<h3>DataCite</h3>{convert(self.dc)}'
+        return buf
 
     def publish(self, foundry_metadata, data_source, title, authors, update=False,
                 publication_year=None, **kwargs,):
@@ -308,6 +324,7 @@ class Foundry(FoundryMetadata):
             publication_year (int): Year of dataset publication. If None, will
                 be set to the current calendar year by MDF Connect Client.
                 (default: $current_year)
+
         Keyword Args:
             affiliations (list): List of author affiliations
             tags (list): List of tags to apply to the data package
@@ -320,7 +337,6 @@ class Foundry(FoundryMetadata):
             of dataset. Contains `source_id`, which can be used to check the
             status of the submission
         """
-
         self.connect_client.create_dc_block(
             title=title,
             authors=authors,
@@ -329,8 +345,9 @@ class Foundry(FoundryMetadata):
             publisher=kwargs.get("publisher", ""),
             publication_year=publication_year,
         )
-        self.connect_client.add_organization("Foundry")
-        self.connect_client.set_project_block("foundry", foundry_metadata)
+        self.connect_client.add_organization(self.config.organization)
+        self.connect_client.set_project_block(
+            self.config.metadata_key, foundry_metadata)
         self.connect_client.add_data_source(data_source)
         self.connect_client.set_source_name(kwargs.get("short_name", title))
 
@@ -349,7 +366,8 @@ class Foundry(FoundryMetadata):
             title (req)
             authors (req)
             short_name (req)
-            servable_type (req) ("static method", "class method", "keras", "pytorch", "tensorflow", "sklearn")
+            servable_type (req) ("static method", "class method",
+                           "keras", "pytorch", "tensorflow", "sklearn")
             affiliations
             domains
             abstract
@@ -383,15 +401,22 @@ class Foundry(FoundryMetadata):
         if options["servable"]["type"] == "sklearn":
             model_info = ScikitLearnModel.create_model(options["servable"]["filepath"],
                                                        options["servable"]["n_input_columns"],
-                                                       options["servable"].get("classes", None),
-                                                       options["servable"].get("serialization_method", "pickle")
+                                                       options["servable"].get(
+                                                           "classes", None),
+                                                       options["servable"].get(
+                                                           "serialization_method", "pickle")
                                                        )
         # TODO: fix weird M1 error with TF
         elif options["servable"]['type'] == "keras":
             model_info = KerasModel.create_model(options["servable"]["model_path"],
-                                                 options["servable"].get("output_names", None),
-                                                 options["servable"].get("arch_path", None),
-                                                 options["servable"].get("custom_objects", None)
+                                                 options["servable"].get(
+                                                     "output_names", None),
+                                                 options["servable"].get(
+                                                     "arch_path", None),
+                                                 options["servable"].get(
+                                                     "custom_objects", None),
+                                                 options["servable"].get(
+                                                     "force_tf_keras", False)
                                                  )
         else:
             raise ValueError("Servable type '{}' is not recognized, please use one of the following types: \n"
@@ -406,7 +431,8 @@ class Foundry(FoundryMetadata):
         model_info.set_name(options["short_name"])
         model_info.set_title(options["title"])
         # TODO: fix bug where if you put in name without comma, get list index out of range error
-        model_info.set_authors(options["authors"], options.get("affiliations", []))
+        model_info.set_authors(
+            options["authors"], options.get("affiliations", []))
         # TODO: dont pass in {} as default, overwrites everything -- should def document this
         # TODO: consider whether that's desired functionality (should users be able to specify 1 or 2 requirements, but
         #   the container still has other pre-loaded dependencies?
@@ -511,8 +537,24 @@ class Foundry(FoundryMetadata):
         """
         # Check if the dir already exists
         path = os.path.join(self.config.local_cache_dir, self.mdf["source_id"])
+
         if os.path.isdir(path):
-            return self
+            # if directory is present, but doesn't have the correct number of files inside, dataset will attempt to redownload
+            dir_length = len(os.listdir(path))
+            if self.dataset.splits:
+                # if metadata indicates splits, check that the directory has as many files as there are splits
+                if(dir_length == len(self.dataset.splits)):
+                    return self
+                else:
+                    print(
+                        "Unexpected number of files in directory -- dataset will be redownloaded.")
+            else:
+                # in the case of no splits, ensure the directory contains the data file
+                if(dir_length == 1):
+                    return self
+                else:
+                    print(
+                        "Unexpected number of files in directory -- dataset will be redownloaded.")
 
         res = self.forge_client.search(
             "mdf.source_id:{name}".format(name=self.mdf["source_id"]), advanced=True
@@ -527,132 +569,30 @@ class Foundry(FoundryMetadata):
             )
         else:
             source_id = self.mdf["source_id"]
-            xtract_base_url = (
-                "http://xtract-crawler-4.eba-ghixpmdf.us-east-1.elasticbeanstalk.com"
-            )
-
-            # MDF Materials Data at NCSA
-            source_ep_id = "82f1b5c6-6e9b-11e5-ba47-22000b92c6ec"
-            base_url = "https://data.materialsdatafacility.org"
-            folder_to_crawl = f"/foundry/{source_id}/"
-
-            # This only matters if you want files grouped together.
-            grouper = "matio"
-
-            auth_token = self.xtract_tokens["auth_token"]
-            transfer_token = self.xtract_tokens["transfer_token"]
-            funcx_token = self.xtract_tokens["funcx_token"]
-
-            headers = {
-                "Authorization": auth_token,
-                "Transfer": transfer_token,
-                "FuncX": funcx_token,
-                "Petrel": auth_token,
+            xtract_config = {
+                "xtract_base_url": "http://xtract-crawler-4.eba-ghixpmdf.us-east-1.elasticbeanstalk.com",
+                "source_ep_id": "82f1b5c6-6e9b-11e5-ba47-22000b92c6ec",
+                "base_url": "https://data.materialsdatafacility.org",
+                "folder_to_crawl": f"/foundry/{source_id}/",
+                "grouper": "matio"
             }
-            if verbose:
-                print(f"Headers: {headers}")
+            xtract_https_download(self, verbose=verbose, **xtract_config)
 
-            # Initialize the crawl. This kicks off the Globus EP crawling service on the backend.
-            crawl_url = f"{xtract_base_url}/crawl"
-            if verbose:
-                print(f"Crawl URL is : {crawl_url}")
-
-            first_ep_dict = {
-                "repo_type": "GLOBUS",
-                "eid": source_ep_id,
-                "dir_paths": [folder_to_crawl],
-                "grouper": grouper,
-            }
-            tokens = {"Transfer": transfer_token, "Authorization": funcx_token}
-            crawl_req = requests.post(
-                f"{xtract_base_url}/crawl",
-                json={"endpoints": [first_ep_dict], "tokens": tokens},
-            )
-
-            if verbose:
-                print("Crawl response:", crawl_req)
-            crawl_id = json.loads(crawl_req.content)["crawl_id"]
-            if verbose:
-                print(f"Crawl ID: {crawl_id}")
-
-            # Wait for the crawl to finish before we can start fetching our metadata.
-            while True:
-                crawl_status = requests.get(
-                    f"{xtract_base_url}/get_crawl_status", json={"crawl_id": crawl_id}
-                )
-                if verbose:
-                    print(crawl_status)
-                crawl_content = json.loads(crawl_status.content)
-                if verbose:
-                    print(f"Crawl Status: {crawl_content}")
-
-                if crawl_content["crawl_status"] == "complete":
-                    files_crawled = crawl_content["files_crawled"]
-                    if verbose:
-                        print("Our crawl has succeeded!")
-                    break
-                else:
-                    if verbose:
-                        print("Sleeping before re-polling...")
-                    time.sleep(2)
-
-            # Now we fetch our metadata. Here you can configure n to be maximum number of
-            # messages you want at once.
-
-            file_ls = []
-            fetched_files = 0
-            while fetched_files < files_crawled:
-                fetch_mdata = requests.get(
-                    f"{xtract_base_url}/fetch_crawl_mdata",
-                    json={"crawl_id": crawl_id, "n": 2},
-                )
-                fetch_content = json.loads(fetch_mdata.content)
-
-                for file_path in fetch_content["file_ls"]:
-                    file_ls.append(file_path)
-                    fetched_files += 1
-
-                if fetch_content["queue_empty"]:
-                    if verbose:
-                        print("Queue is empty! Continuing...")
-                    time.sleep(2)
-
-            source_path = os.path.join(
-                self.config.local_cache_dir, self.mdf["source_id"]
-            )
-
-            if not os.path.exists(self.config.local_cache_dir):
-                os.mkdir(self.config.local_cache_dir)
-                os.mkdir(source_path)
-
-            elif not os.path.exists(source_path):
-                os.mkdir(source_path)
-
-            num_cores = multiprocessing.cpu_count()
-
-            def download_file(file):
-                requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
-
-                url = "https://data.materialsdatafacility.org" + file["path"]
-                destination = (
-                    "data/"
-                    + source_id
-                    + "/"
-                    + file["path"][file["path"].rindex("/") + 1 :]
-                )
-                response = requests.get(url, verify=False)
-
-                with open(destination, "wb") as f:
-                    f.write(response.content)
-
-                return {file["path"] + " status": True}
-
-            results = Parallel(n_jobs=num_cores)(
-                delayed(download_file)(file) for file in file_ls
-            )
-
-            print("Done curling.")
-            print(results)
+        # after download check making sure directory exists, contains proper amount of files
+        if os.path.isdir(path):
+            # checking for proper number of files downloaded
+            dir_length = len(os.listdir(path))
+            if self.dataset.splits:
+                if(dir_length != len(self.dataset.splits)):
+                    raise FileNotFoundError(
+                        "Incorrect number of files in download directory")
+            else:
+                if(dir_length != 1):
+                    raise FileNotFoundError(
+                        "Incorrect number of files in download directory")
+        else:
+            raise NotADirectoryError(
+                "Unable to create directory to download data")
 
         return self
 
@@ -668,6 +608,14 @@ class Foundry(FoundryMetadata):
         -------
         (Foundry): self: for chaining
         """
+        if as_object:
+            return [key for key in self.dataset.keys if key.type == type]
+        else:
+            keys = [key.key for key in self.dataset.keys if key.type == type]
+            key_list = []
+            for k in keys:
+                key_list = key_list + k
+            return key_list
 
         print("Building Data Package")
         num_cores = multiprocessing.cpu_count()
@@ -692,3 +640,129 @@ class Foundry(FoundryMetadata):
         )
 
         return self
+
+    def get_keys(self, type=None, as_object=False):
+        """Get keys for a Foundry dataset
+
+        Arguments:
+            type (str): The type of key to be returned e.g., "input", "target"
+            as_object (bool): When ``False``, will return a list of keys in as strings
+                    When ``True``, will return the full key objects
+                    **Default:** ``False``
+        Returns: (list) String representations of keys or if ``as_object``
+                    is False otherwise returns the full key objects.
+
+        """
+
+        if as_object:
+            if type:
+                return [key for key in self.dataset.keys if key.type == type]
+            else:
+                return [key for key in self.dataset.keys]
+
+        else:
+            if type:
+                keys = [key.key for key in self.dataset.keys if key.type == type]
+            else:
+                keys = [key.key for key in self.dataset.keys]
+
+            key_list = []
+            for k in keys:
+                key_list = key_list + k
+            return key_list
+
+
+    def _load_data(self, file=None, source_id=None, globus=True):
+
+        # Build the path to access the cached data
+        if source_id:
+            path = os.path.join(self.config.local_cache_dir, source_id)
+        else:
+            path = os.path.join(self.config.local_cache_dir,
+                                self.mdf["source_id"])
+
+        # Handle Foundry-defined types.
+        if self.dataset.data_type.value == "tabular":
+            # Determine which file to load, defaults to config.dataframe_file
+            if not file:
+                file = self.config.dataframe_file
+
+            # Check to make sure the path can be created
+            try:
+                path_to_file = os.path.join(path, file)
+            except Exception as e:
+                print("Unable to find path to file for download: {}".format(e))
+                raise e
+
+            # Check to see whether file exists at path
+            if not os.path.isfile(path_to_file):
+                raise FileNotFoundError(
+                    "No file found at expected path: {}".format(path_to_file))
+            # If the file is not local, fetch the contents with Globus
+            # Check if the contents are local
+            # TODO: Add hashes and versioning to metadata and checking to the file
+            try:
+                self.dataset.dataframe = pd.read_json(
+                    path_to_file
+                )
+            except Exception as e:
+                print("Reading {} as JSON failed: {} \n".format(
+                    file, e), "Now attempting to read as JSONL")
+                try:
+                    # Try to read individual lines instead
+                    self.dataset.dataframe = pd.read_json(
+                        path_to_file, lines=True
+                    )
+                except Exception as f:
+                    print("Reading {} as JSONL failed: {} \n".format(
+                        file, f), "Now attempting to read as CSV")
+                    try:
+                        # Try to read as CSV instead
+                        self.dataset.dataframe = pd.read_csv(
+                            path_to_file
+                        )
+
+                    except Exception as g:
+                        print(
+                            "Reading {} as CSV failed, unable to load data properly: {}".format(file, g))
+                        raise e
+
+            return (
+                self.dataset.dataframe[self.get_keys("input")],
+                self.dataset.dataframe[self.get_keys("target")],
+            )
+
+        elif self.dataset.data_type.value == "hdf5":
+            if not file:
+                file = self.config.data_file
+            f = h5py.File(os.path.join(path, file), "r")
+
+            special_types = ["input", "target"]
+            tmp_data = {s: {} for s in special_types}
+            for s in special_types:
+                for key in self.get_keys(s):
+                    if isinstance(f[key], h5py.Group):
+                        if is_pandas_pytable(f[key]):
+                            df = pd.read_hdf(file, key)
+                            tmp_data[s][key] = df
+                        else:
+                            tmp_data[s][key] = f[key]
+                    elif isinstance(f[key], h5py.Dataset):
+                        tmp_data[s][key] = f[key][0:]
+            return tmp_data
+        else:
+            raise NotImplementedError
+
+
+def is_pandas_pytable(group):
+    if 'axis0' in group.keys() and 'axis1' in group.keys():
+        return True
+    else:
+        return False
+
+
+def is_doi(string: str):
+    if string.startswith('10.') or string.startswith('https://doi.org/'):
+        return True
+    else:
+        return False
